/*
 * classic_visitor.cpp
 *
 *  Created on: Mar 25, 2011
 *      Author: lipari
 */

#include "classic_visitor.hpp"
#include <iostream>
#include <string>

namespace classic_visitor
{
	using namespace std;

	class Visitor;

	class Node {
	protected:
		string type;
	public:
		Node (const string &n) : type(n) {}
		virtual ~Node() {}

		string getType() const { return type; }
		virtual bool lvalue() const = 0;
		virtual bool isLeaf() { return true; }

		virtual void accept(Visitor *v) = 0;
	};

	class CompositeNode : public Node {
	public:
		CompositeNode(const string &n) : Node(n) {}
		bool isLeaf() { return false; }

		virtual Node * getLeft() = 0;
		virtual Node * getRight() = 0;
	};

	class VariableRefNode : public Node {
		string name;
	public:
		VariableRefNode(const string &n) : Node("var"), name(n) {}
		bool lvalue() const { return true; }
		string getName() { return name; }

		void accept(Visitor *v);
	};

	class NumberNode : public Node {
		int number;
	public:
		NumberNode(int i) : Node("number"), number(i) {}
		bool lvalue() const { return false; }
		int getValue() { return number; }

		void accept(Visitor *v);
	};

	class BinaryOperatorNode : public CompositeNode {
		Node *left;
		Node *right;
		string op;
	public:
		BinaryOperatorNode(const string &opname, const string &n) :
			CompositeNode(opname), left(0), right(0), op(n) {}

		void setLeft(Node *v) { left = v; }
		void setRight(Node *v) { right = v; }
		bool lvalue() const { return false; }
		Node * getLeft() { return left; }
		Node * getRight() { return right; }
	};

	class AssignmentNode : public BinaryOperatorNode {
	public:
		AssignmentNode() : BinaryOperatorNode("assignment", "=") {}

		void accept(Visitor *v);

	};

	class SumNode : public BinaryOperatorNode {
	public:
		SumNode() : BinaryOperatorNode("sum", "+") {}

		void accept(Visitor *v);
	};

	/*-----------------------------*/

	// visitor hierarchy
	class Visitor {
	protected:
		~Visitor() {}
	public:
		virtual void visitVarRefNode(VariableRefNode *n) = 0;
		virtual void visitNumberNode(NumberNode *n) = 0;
		virtual void visitAssignmentNode(AssignmentNode *n) = 0;
		virtual void visitSumNode(SumNode *n) = 0;
	};

	void VariableRefNode::accept(Visitor *v) {
		v->visitVarRefNode(this);
	}

	void NumberNode::accept(Visitor *v) {
		v->visitNumberNode(this);
	}

	void AssignmentNode::accept(Visitor *v) {
		v->visitAssignmentNode(this);
	}

	void SumNode::accept(Visitor *v) {
		v->visitSumNode(this);
	}

	// Visitors
	class GenerateCode : public Visitor {
	public:
		GenerateCode() {}

		void visitVarRefNode(VariableRefNode *n) {
			cout << n->getName();
		}
		void visitNumberNode(NumberNode *n) {
			cout << n->getValue();
		}
		void visitAssignmentNode(AssignmentNode *n) {
			cout << " = ";
		}
		void visitSumNode(SumNode *n) {
			cout << " + ";
		}
	};

	class TypeCheck : public Visitor {
		bool error;
	public:
		TypeCheck() : error(false) {}

		bool isError() const { return error; }

		void visitVarRefNode(VariableRefNode *n) {
			cout << n->getName() << "ok" << endl;
		}
		void visitNumberNode(NumberNode *n) {
			cout << n->getValue() << "ok" << endl;
		}
		void visitAssignmentNode(AssignmentNode *n) {
			Node *l = n->getLeft();
			if (l==0) {
				cout << "Error, missing left value in assignment" << endl;
				error = true;
			} else if (!l->lvalue()) {
				cout << "Error, " << l->getType() << " is not a lvalue in assigment" << endl;
				error = true;
			}
		}

		void visitSumNode(SumNode *n) {
			cout << n->getType() << "ok";
		}
	};

	void in_order_visit(Node *n, Visitor *v)
	{
		if (n->isLeaf()) n->accept(v);
		else {
			CompositeNode *cn = dynamic_cast<CompositeNode *>(n);
			if (cn == 0) throw "Error, should be a composite";
			if (cn->getLeft()) in_order_visit(cn->getLeft(), v);
			cn->accept(v);
			if (cn->getRight()) in_order_visit(cn->getRight(), v);
		}
	}

	void gen_code(Node *n)
	{
		GenerateCode visitor;
		in_order_visit(n, &visitor);
	}

	void type_check(Node *n)
	{
		TypeCheck visitor;
		in_order_visit(n, &visitor);
		if (visitor.isError()) cout << "An error occurred" << endl;
	}

	int run()
	{
		// prepare a set of nodes, first make a correct one
		AssignmentNode an;
		VariableRefNode vr("x");
		NumberNode n1(10);
		NumberNode n2(20);
		SumNode sn;

		sn.setLeft(&n1);
		sn.setRight(&n2);

		an.setLeft(&sn);
		an.setRight(&vr);

		type_check(&an);
		gen_code(&an);

		return 1;
	}
}

